package sun.nio.fs;

import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.file.StandardOpenOption;
import java.util.HashSet;
import java.io.IOException;
import java.util.Collections;
import java.util.ArrayList;
import java.util.List;
import sun.misc.Unsafe;

class WindowsUserDefinedFileAttributeView extends AbstractUserDefinedFileAttributeView
{
    private static final Unsafe unsafe;
    private final WindowsPath file;
    private final boolean followLinks;
    
    private String join(String str, String str2) {
        if (str2 == null) {
            throw new NullPointerException("'name' is null");
        }
        return str + ":" + str2;
    }
    
    private String join(WindowsPath windowsPath, String s) throws WindowsException {
        return this.join(windowsPath.getPathForWin32Calls(), s);
    }
    
    WindowsUserDefinedFileAttributeView(WindowsPath file, boolean followLinks) {
        this.file = file;
        this.followLinks = followLinks;
    }
    
    private List<String> listUsingStreamEnumeration() throws IOException {
        ArrayList<Object> list = new ArrayList();
        try {
            WindowsNativeDispatcher.FirstStream findFirstStream = WindowsNativeDispatcher.FindFirstStream(this.file.getPathForWin32Calls());
            if (findFirstStream != null) {
                long handle = findFirstStream.handle();
                try {
                    String name = findFirstStream.name();
                    if (!name.equals("::$DATA")) {
                        list.add(name.split(":")[1]);
                    }
                    String findNextStream;
                    while ((findNextStream = WindowsNativeDispatcher.FindNextStream(handle)) != null) {
                        list.add(findNextStream.split(":")[1]);
                    }
                }
                finally {
                    WindowsNativeDispatcher.FindClose(handle);
                }
            }
        }
        catch (WindowsException ex) {
            ex.rethrowAsIOException(this.file);
        }
        return Collections.unmodifiableList(list);
    }
    
    private List<String> listUsingBackupRead() throws IOException {
        long createFile = -1L;
        try {
            int n = 33554432;
            if (!this.followLinks && this.file.getFileSystem().supportsLinks()) {
                n |= 0x200000;
            }
            createFile = WindowsNativeDispatcher.CreateFile(this.file.getPathForWin32Calls(), Integer.MIN_VALUE, 1, 3, n);
        }
        catch (WindowsException ex) {
            ex.rethrowAsIOException(this.file);
        }
        NativeBuffer nativeBuffer = null;
        ArrayList list = new ArrayList();
        try {
            nativeBuffer = NativeBuffers.getNativeBuffer(4096);
            long address = nativeBuffer.address();
            long context = 0L;
            while (true) {
                try {
                    WindowsNativeDispatcher.BackupResult backupRead = WindowsNativeDispatcher.BackupRead(createFile, address, 20, false, context);
                    context = backupRead.context();
                    if (backupRead.bytesTransferred() != 0) {
                        int int1 = WindowsUserDefinedFileAttributeView.unsafe.getInt(address + 0L);
                        long long1 = WindowsUserDefinedFileAttributeView.unsafe.getLong(address + 8L);
                        int int2 = WindowsUserDefinedFileAttributeView.unsafe.getInt(address + 16L);
                        if (int2 <= 0 || WindowsNativeDispatcher.BackupRead(createFile, address, int2, false, context).bytesTransferred() == int2) {
                            if (int1 == 4) {
                                char[] array = new char[int2 / 2];
                                WindowsUserDefinedFileAttributeView.unsafe.copyMemory(null, address, array, (long)Unsafe.ARRAY_CHAR_BASE_OFFSET, (long)int2);
                                String[] split = new String(array).split(":");
                                if (split.length == 3) {
                                    list.add(split[1]);
                                }
                            }
                            if (int1 == 9) {
                                throw new IOException("Spare blocks not handled");
                            }
                            if (long1 > 0L) {
                                WindowsNativeDispatcher.BackupSeek(createFile, long1, context);
                            }
                            else {
                                continue;
                            }
                            continue;
                        }
                    }
                }
                catch (WindowsException ex2) {
                    throw new IOException(ex2.errorString());
                }
                finally {
                    if (context != 0L) {
                        try {
                            WindowsNativeDispatcher.BackupRead(createFile, 0L, 0, true, context);
                        }
                        catch (WindowsException ex3) {}
                    }
                }
            }
        }
        finally {
            if (nativeBuffer != null) {
                nativeBuffer.release();
            }
            WindowsNativeDispatcher.CloseHandle(createFile);
        }
        return Collections.unmodifiableList(list);
    }
    
    @Override
    public List<String> list() throws IOException {
        if (System.getSecurityManager() != null) {
            this.checkAccess(this.file.getPathForPermissionCheck(), true, false);
        }
        if (this.file.getFileSystem().supportsStreamEnumeration()) {
            return this.listUsingStreamEnumeration();
        }
        return this.listUsingBackupRead();
    }
    
    @Override
    public int size(String s) throws IOException {
        if (System.getSecurityManager() != null) {
            this.checkAccess(this.file.getPathForPermissionCheck(), true, false);
        }
        FileChannel fileChannel = null;
        try {
            HashSet set = new HashSet();
            set.add(StandardOpenOption.READ);
            if (!this.followLinks) {
                set.add(WindowsChannelFactory.OPEN_REPARSE_POINT);
            }
            fileChannel = WindowsChannelFactory.newFileChannel(this.join(this.file, s), null, set, 0L);
        }
        catch (WindowsException ex) {
            ex.rethrowAsIOException(this.join(this.file.getPathForPermissionCheck(), s));
        }
        try {
            long size = fileChannel.size();
            if (size > 2147483647L) {
                throw new ArithmeticException("Stream too large");
            }
            return (int)size;
        }
        finally {
            fileChannel.close();
        }
    }
    
    @Override
    public int read(String s, ByteBuffer byteBuffer) throws IOException {
        if (System.getSecurityManager() != null) {
            this.checkAccess(this.file.getPathForPermissionCheck(), true, false);
        }
        FileChannel fileChannel = null;
        try {
            HashSet set = new HashSet();
            set.add(StandardOpenOption.READ);
            if (!this.followLinks) {
                set.add(WindowsChannelFactory.OPEN_REPARSE_POINT);
            }
            fileChannel = WindowsChannelFactory.newFileChannel(this.join(this.file, s), null, set, 0L);
        }
        catch (WindowsException ex) {
            ex.rethrowAsIOException(this.join(this.file.getPathForPermissionCheck(), s));
        }
        try {
            if (fileChannel.size() > (long)byteBuffer.remaining()) {
                throw new IOException("Stream too large");
            }
            int n = 0;
            while (byteBuffer.hasRemaining()) {
                int read = fileChannel.read(byteBuffer);
                if (read < 0) {
                    break;
                }
                n += read;
            }
            return n;
        }
        finally {
            fileChannel.close();
        }
    }
    
    @Override
    public int write(String s, ByteBuffer byteBuffer) throws IOException {
        if (System.getSecurityManager() != null) {
            this.checkAccess(this.file.getPathForPermissionCheck(), false, true);
        }
        long createFile = -1L;
        try {
            int n = 33554432;
            if (!this.followLinks) {
                n |= 0x200000;
            }
            createFile = WindowsNativeDispatcher.CreateFile(this.file.getPathForWin32Calls(), Integer.MIN_VALUE, 7, 3, n);
        }
        catch (WindowsException ex) {
            ex.rethrowAsIOException(this.file);
        }
        try {
            HashSet o = new HashSet();
            if (!this.followLinks) {
                o.add(WindowsChannelFactory.OPEN_REPARSE_POINT);
            }
            o.add(StandardOpenOption.CREATE);
            o.add(StandardOpenOption.WRITE);
            o.add(StandardOpenOption.TRUNCATE_EXISTING);
            FileChannel fileChannel = null;
            try {
                fileChannel = WindowsChannelFactory.newFileChannel(this.join(this.file, s), null, o, 0L);
            }
            catch (WindowsException ex2) {
                ex2.rethrowAsIOException(this.join(this.file.getPathForPermissionCheck(), s));
            }
            try {
                int remaining = byteBuffer.remaining();
                while (byteBuffer.hasRemaining()) {
                    fileChannel.write(byteBuffer);
                }
                fileChannel.close();
                return remaining;
            }
            finally {
                fileChannel.close();
            }
        }
        finally {
            WindowsNativeDispatcher.CloseHandle(createFile);
        }
    }
    
    @Override
    public void delete(String s) throws IOException {
        if (System.getSecurityManager() != null) {
            this.checkAccess(this.file.getPathForPermissionCheck(), false, true);
        }
        String join = this.join(WindowsLinkSupport.getFinalPath(this.file, this.followLinks), s);
        try {
            WindowsNativeDispatcher.DeleteFile(join);
        }
        catch (WindowsException ex) {
            ex.rethrowAsIOException(join);
        }
    }
    
    static {
        unsafe = Unsafe.getUnsafe();
    }
}
